# @file wincert.py
# Code to work with UEFI WinCert data
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

"""Module for working with UEFI WinCert data."""


import io
import struct
import uuid
from edk2toollib.utility_functions import PrintByteList


class WinCertPkcs1(object):
    """Object representing a WinCertPkcs1 struct.

    Certificate which encapsulates the RSASSA_PKCS1-v1_5 digital signature.

    Attributes:
        Hdr_dwLength (int):         The length of the entire certificate, including the length of the header, in bytes.
        Hdr_wRevision (int):        The certificate type. See WIN_CERT_TYPE_xxx for the UEFI certificate types. The UEFI
                                    specification reserves the range of certificate type values from 0x0EF0 to 0x0EFF.
        Hdr_wCertificateType (int): The following is the actual certificate. The format of the certificate depends on
                                    wCertificateType.
        HashAlgorithm (uuid.UUID):  The Guid representing the hash algorithm for the Cert.
        CertData (memoryview):      The actual Cert.

    The WIN_CERTIFICATE_UEFI_PKCS1_15 structure is derived from
    WIN_CERTIFICATE and encapsulate the information needed to
    implement the RSASSA-PKCS1-v1_5 digital signature algorithm as
    specified in RFC2437.
    typedef struct {
        WIN_CERTIFICATE Hdr;
        EFI_GUID        HashAlgorithm;
        UINT8 Signature[];
    } WIN_CERTIFICATE_EFI_PKCS1_15;

    The WIN_CERTIFICATE structure is part of the PE/COFF specification.
    typedef struct {
        UINT32  dwLength;
        UINT16  wRevision;
        UINT16  wCertificateType;
        UINT8 bCertificate[ANYSIZE_ARRAY];
    } WIN_CERTIFICATE;
    """
    STATIC_STRUCT_SIZE = (4 + 2 + 2 + 16)
    EFI_HASH_SHA256 = uuid.UUID("{51AA59DE-FDF2-4EA3-BC63-875FB7842EE9}")  # EFI_HASH_SHA256 guid defined by UEFI spec

    def __init__(self, filestream=None):
        """Inits the object."""
        if (filestream is None):
            self.Hdr_dwLength = WinCertPkcs1.STATIC_STRUCT_SIZE
            self.Hdr_wRevision = WinCert.REVISION
            self.Hdr_wCertificateType = WinCert.WIN_CERT_TYPE_EFI_PKCS115
            self.HashAlgorithm = None
            self.CertData = None
        else:
            self.PopulateFromFileStream(filestream)

    def AddCertData(self, fs):
        """Adds the Cert Data to the struct.

        Raises:
            (Exception): Invalid Cert Data
            (Exception): Missing Hash Algorithm
        """
        if (self.CertData is not None):
            raise Exception("Cert Data not 0")
        if (self.HashAlgorithm is None):
            raise Exception("You must set the Hash Algorithm first")
        self.CertData = fs.read()
        self.Hdr_dwLength = self.Hdr_dwLength + len(self.CertData)

    def PopulateFromFileStream(self, fs):
        """Populates the struct from a filestream.

        Args:
            fs (obj): An open file

        Raises:
            (Exception): Invalid stream
            (Exception): Invalid stream size
        """
        if (fs is None):
            raise Exception("Invalid File stream")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if ((end - offset) < WinCertPkcs1.STATIC_STRUCT_SIZE):  # size of the static header data
            raise Exception("Invalid file stream size")

        self.Hdr_dwLength = struct.unpack("=I", fs.read(4))[0]
        self.Hdr_wRevision = struct.unpack("=H", fs.read(2))[0]
        self.Hdr_wCertificateType = struct.unpack("=H", fs.read(2))[0]
        self.HashAlgorithm = uuid.UUID(bytes_le=fs.read(16))
        self.CertData = None

        if ((end - fs.tell()) < 1):
            raise Exception("Invalid File stream. No data for signature cert data")

        if ((end - fs.tell()) < (self.Hdr_dwLength - WinCertPkcs1.STATIC_STRUCT_SIZE)):
            raise Exception("Invalid file stream size")

        self.CertData = memoryview(fs.read(self.Hdr_dwLength - WinCertPkcs1.STATIC_STRUCT_SIZE))

    def Print(self):
        """Prints the struct to the console."""
        print("WinCertPKCS115")
        print("  Hdr_dwLength:         0x%X" % self.Hdr_dwLength)
        print("  Hdr_wRevision:        0x%X" % self.Hdr_wRevision)
        print("  Hdr_wCertificateType: 0x%X" % self.Hdr_wCertificateType)
        print("  Hash Guid:            %s" % str(self.HashAlgorithm))
        print("  CertData:             ")
        cdl = self.CertData.tolist()
        PrintByteList(cdl)

    def Write(self, fs):
        r"""Serializes the object.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        fs.write(struct.pack("=I", self.Hdr_dwLength))
        fs.write(struct.pack("=H", self.Hdr_wRevision))
        fs.write(struct.pack("=H", self.Hdr_wCertificateType))
        fs.write(self.HashAlgorithm.bytes_le)
        fs.write(self.CertData)


class WinCertUefiGuid(object):
    """Object representing a Certificate which encapsulates a GUID-specific digital signature.

    Attributes:
        Hdr_dwLength (int):         The length of the entire certificate, including the length of the header, in bytes.
        Hdr_wRevision (int):        The certificate type. See WIN_CERT_TYPE_xxx for the UEFI certificate types. The UEFI
                                    specification reserves the range of certificate type values from 0x0EF0 to 0x0EFF.
        Hdr_wCertificateType (int): The following is the actual certificate. The format of the certificate depends on
                                    wCertificateType.
        CertData (memoryview):      The actual Cert.

    typedef struct {
        WIN_CERTIFICATE   Hdr;
        EFI_GUID          CertType;
        UINT8            CertData[1];
    } WIN_CERTIFICATE_UEFI_GUID;

    The WIN_CERTIFICATE structure is part of the PE/COFF specification.
    typedef struct {
        UINT32  dwLength;
        UINT16  wRevision;
        UINT16  wCertificateType;
        UINT8 bCertificate[ANYSIZE_ARRAY];
    } WIN_CERTIFICATE;
    """

    _StructFormat = '<IHH16s'
    _StructSize = struct.calcsize(_StructFormat)

    _EFI_CERT_TYPE_PKCS7_GUID = uuid.UUID('4aafd29d-68df-49ee-8aa9-347d375665a7')

    # Preserved for back compat.
    STATIC_STRUCT_SIZE = _StructSize
    PKCS7Guid = _EFI_CERT_TYPE_PKCS7_GUID

    def __init__(self, in_data=None):
        """Inits the object."""
        self.Hdr_dwLength = self._StructSize
        self.Hdr_wRevision = WinCert.REVISION
        self.Hdr_wCertificateType = WinCert.WIN_CERT_TYPE_EFI_GUID
        self.CertType = self._EFI_CERT_TYPE_PKCS7_GUID
        self.CertData = b''

        if in_data is not None:
            # Account for back compat. Behave differently for file streams.
            if hasattr(in_data, 'seek'):
                self.PopulateFromFileStream(in_data)
            else:
                self.Decode(in_data)

    def Encode(self):
        r"""Serializes the object.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')

        Raises:
            (ValueError): Invalid Revision
            (ValueError): Invalid Cert Type
        """
        if self.Hdr_wRevision != WinCert.REVISION:
            raise ValueError
        if self.Hdr_wCertificateType != WinCert.WIN_CERT_TYPE_EFI_GUID:
            raise ValueError
        if self.CertType != self._EFI_CERT_TYPE_PKCS7_GUID:
            raise ValueError
        self.Hdr_dwLength = self._StructSize + len(self.CertData)

        WinCertHeader = struct.pack(
            self._StructFormat,
            self.Hdr_dwLength,
            self.Hdr_wRevision,
            self.Hdr_wCertificateType,
            self.CertType.bytes_le
        )

        return WinCertHeader + self.CertData

    def Decode(self, Buffer):
        """Loads the struct with values from a buffer.

        Args:
            Buffer (obj): Buffer containing serialized data

        Returns:
            (obj): Any remaining buffer

        Raises:
            (ValueError): Invalid Buffer
            (ValueError): Invalid dw length
            (ValueError): Invalid Revision
            (ValueError): Invalid Cert Type
        """
        if len(Buffer) < self._StructSize:
            raise ValueError
        (dwLength, wRevision, wCertificateType, CertType) = struct.unpack(
            self._StructFormat,
            Buffer[0:self._StructSize]
        )
        if dwLength < self._StructSize:
            raise ValueError
        if wRevision != WinCert.REVISION:
            raise ValueError
        if wCertificateType != WinCert.WIN_CERT_TYPE_EFI_GUID:
            raise ValueError
        if CertType != self._EFI_CERT_TYPE_PKCS7_GUID.bytes_le:
            raise ValueError
        self.Hdr_dwLength = dwLength
        self.Hdr_wRevision = wRevision
        self.Hdr_wCertificateType = wCertificateType
        self.CertType = uuid.UUID(bytes_le=CertType)
        self.CertData = Buffer[self._StructSize:self.Hdr_dwLength]

        # Return the remaining buffer, if any exists.
        return Buffer[self.Hdr_dwLength:]

    def AddCertData(self, in_data):
        """Adds the Cert Data to the struct.

        Args:
            in_data (obj): Data to read.
        """
        # Account for back compat. Behave differently for file streams.
        if hasattr(in_data, 'seek'):
            self.CertData = in_data.read()
        else:
            self.CertData = in_data

        self.Hdr_dwLength = self.Hdr_dwLength + len(self.CertData)

    def PopulateFromFileStream(self, fs):
        """Un-serialized from a filestream.

        Args:
            fs (obj): Already opened file

        Raises:
            (ValueError): Invalid fs
            (ValueError): Invalid size
        """
        if fs is None:
            raise ValueError

        # Determine the end of the stream.
        current = fs.tell()
        end = fs.seek(0, io.SEEK_END)
        fs.seek(current)

        # Make sure that we can at least parse the size field.
        field_string = "<I"
        field_size = struct.calcsize(field_string)
        if (end - current) < field_size:
            raise ValueError

        # Parse the size field.
        (buffer_size,) = struct.unpack(field_string, fs.read(field_size))

        if (end - current) < buffer_size:
            raise ValueError

        fs.seek(current)
        object_buffer = fs.read(buffer_size)

        return self.Decode(object_buffer)

    def Print(self):
        """Prints struct to console."""
        self.DumpInfo()

    def DumpInfo(self):
        """Prints struct to console."""
        print('WIN_CERTIFICATE.dwLength         = {dwLength:08X}'
              .format(dwLength=self.Hdr_dwLength))
        print('WIN_CERTIFICATE.wRevision        = {wRevision:04X}'
              .format(wRevision=self.Hdr_wRevision))
        print('WIN_CERTIFICATE.wCertificateType = {wCertificateType:04X}'
              .format(wCertificateType=self.Hdr_wCertificateType))
        print('WIN_CERTIFICATE_UEFI_GUID.CertType             = {Guid}'
              .format(Guid=str(self.CertType).upper()))
        print('sizeof (WIN_CERTIFICATE_UEFI_GUID.CertData)    = {Size:08X}'
              .format(Size=len(self.CertData)))

    def Write(self, fs):
        """Writes the struct to a filestream."""
        fs.write(self.Encode())


class WinCert(object):
    """Object for generating a WinCert."""
    STATIC_STRUCT_SIZE = 8
    # WIN_CERTIFICATE.wCertificateTypes UEFI Spec defined
    WIN_CERT_TYPE_NONE = 0x0000
    WIN_CERT_TYPE_PKCS_SIGNED_DATA = 0x0002
    WIN_CERT_TYPE_EFI_PKCS115 = 0x0EF0
    WIN_CERT_TYPE_EFI_GUID = 0x0EF1
    # Revision
    REVISION = 0x200

    #
    # this method is a factory
    #
    @staticmethod
    def Factory(fs):
        """Generates a specific Cert Type depending on parsed Hdr_wCertificationType from the fs.

        Args:
            fs (obj): filestream

        Returns:
            (WinCertUefiGuid): if Hdr_wCertificationType == WIN_CERT_TYPE_EFI_GUID
            (WinCertPkcs1): if Hdr_wCertificationType == WIN_CERT_TYPE_EFI_PKCS115
        Raises:
            (Exception): Invalid fs
            (Exception): Invalid fs size
        """
        if (fs is None):
            raise Exception("Invalid File stream")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if ((end - offset) < WinCert.STATIC_STRUCT_SIZE):  # size of the static header data
            raise Exception("Invalid file stream size")
        # 1 read len
        # 2 read revision
        # 3 read cert type
        fs.seek(4, 1)  # seeking past Hdr_dwLength
        fs.seek(2, 1)  # seeking past Hdr_wRevision
        Hdr_wCertificateType = struct.unpack("=H", fs.read(2))[0]

        fs.seek(offset)

        if (Hdr_wCertificateType == WinCert.WIN_CERT_TYPE_EFI_GUID):
            return WinCertUefiGuid(fs)
        elif (Hdr_wCertificateType == WinCert.WIN_CERT_TYPE_EFI_PKCS115):
            return WinCertPkcs1(fs)
        else:
            return None



# @file guid_parser_test.py
# Contains unit test routines for the guid parser class.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
from edk2toollib.uefi.edk2.parsers.base_parser import HashFileParser


class TestBaseParser(unittest.TestCase):

    def test_parse_new_section(self):
        parser = HashFileParser("")
        section1 = "[Defines]"
        res, sect = parser.ParseNewSection(section1)
        self.assertTrue(res)
        self.assertEqual(sect, "Defines")
        # invalid section
        section2 = "[Defines"
        res, sect = parser.ParseNewSection(section2)
        self.assertFalse(res)
        # multiple parts with multiple definitions
        section3 = "[Components.X64, Components.IA32]"
        res, sect = parser.ParseNewSection(section3)
        self.assertTrue(res)
        self.assertEqual(sect, "Components")
        # try multiple parts on a single
        section4 = "[ Defines.Common.Section ]"
        res, sect = parser.ParseNewSection(section4)
        self.assertTrue(res)
        self.assertEqual(sect, "Defines")

    def test_strip_comment(self):
        parser = HashFileParser("")
        lines = ["Test", "MagicLib|Include/Magic", ""]
        comments = ["\t#this shouldn't show up", " # test", ""]
        for line in lines:
            for comment in comments:
                test_line = line + comment
                result = parser.StripComment(test_line)
                self.assertEqual(result, line)



##
# UEFI Authenticated Variable Structure Support Library
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module containing Authenticated Variable Structure definitions based on UEFI specification (2.7).

Each object can be created and or populated from a file stream.
Each object can be written to a filesteam as binary and printed to the console in text.
"""
import logging
import datetime
import struct
import hashlib
import uuid
import io
from typing import BinaryIO, List
from operator import attrgetter
from edk2toollib.uefi.wincert import WinCert, WinCertUefiGuid
from edk2toollib.utility_functions import PrintByteList

# spell-checker: ignore decodefs, createfs, deduplicated, deduplication

# UEFI global Variable Namespace
EfiGlobalVarNamespaceUuid = uuid.UUID('8BE4DF61-93CA-11d2-AA0D-00E098032B8C')
Sha256Oid = [0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x01]


class EfiSignatureDataEfiCertX509(object):
    """An object representing a EFI_SIGNATURE_DATA Structure for X509 Certs."""
    STATIC_STRUCT_SIZE = 16
    FIXED_SIZE = False

    def __init__(self,
                 decodefs: BinaryIO = None,
                 decodesize=0,
                 createfs: BinaryIO = None,
                 cert: bytes = None,
                 sigowner: uuid.UUID = None):
        """Inits the object.

        Args:
            decodefs (BinaryIO): a filestream object of binary content that is the structure encoded
            decodesize (int):number of bytes to decode as the EFI_SIGNATURE_DATA object (guid + x509 data)
            createfs (BinaryIO): a filestream object that is the DER encoded x509 cert
            cert (bytes): the x509 certificate to initialize this object
            sigowner (uuid.UUID): the uuid object of the signature owner guid

        Raises:
            (Exception): Invalid FileStream size
            (Exception): Invalid Parameters
        """
        if (decodefs is not None):
            self.PopulateFromFileStream(decodefs, decodesize)
        elif (createfs is not None):
            # create a new one
            self.SignatureOwner = sigowner
            start = createfs.tell()  # should be 0 but maybe this filestream has other things at the head
            createfs.seek(0, 2)
            end = createfs.tell()
            createfs.seek(start)
            self.SignatureDataSize = end - start
            if (self.SignatureDataSize < 0):
                raise Exception("Create File Stream has invalid size")
            self.SignatureData = (createfs.read(self.SignatureDataSize))
        elif (cert is not None):
            self.SignatureOwner = sigowner
            self.SignatureDataSize = len(cert)
            self.SignatureData = cert

        else:
            raise Exception("Invalid Parameters - Not Supported")

    def __lt__(self, other):
        """Less-than comparison for sorting. Looks at SignatureData only, not SignatureOwner."""
        return self.SignatureData < other.SignatureData

    def PopulateFromFileStream(self, fs: BinaryIO, decodesize):
        """Loads an object from a filestream.

        Args:
            fs (BinaryIO): filestream
            decodesize (int): amount to decode.

        Raises:
            (Exception): Invalid filestream
            (Exception): Invalid Decode Size
        """
        if (fs is None):
            raise Exception("Invalid File Steam")

        if (decodesize == 0):
            raise Exception("Invalid Decode Size")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if ((end - offset) < EfiSignatureDataEfiCertX509.STATIC_STRUCT_SIZE):  # size of the  guid
            raise Exception("Invalid file stream size")

        if ((end - offset) < decodesize):  # size requested is too big
            raise Exception("Invalid file stream size vs decodesize")

        self.SignatureOwner = uuid.UUID(bytes_le=fs.read(16))

        # read remaining decode size for x509 data
        self.SignatureDataSize = decodesize - EfiSignatureDataEfiCertX509.STATIC_STRUCT_SIZE
        self.SignatureData = fs.read(self.SignatureDataSize)

    def Print(self, compact: bool = False):
        """Prints to the console."""
        if not compact:
            print("EfiSignatureData - EfiSignatureDataEfiCertX509")
            print("  Signature Owner:      %s" % str(self.SignatureOwner))
            print("  Signature Data: ")
            if (self.SignatureData is None):
                print("    NONE")
            else:
                sdl = self.SignatureData
                if (self.SignatureDataSize != len(sdl)):
                    raise Exception("Invalid Signature Data Size vs Length of data")
                PrintByteList(sdl)
        else:
            s = "ESD:EFI_CERT_X509,"
            s += "%s," % str(self.SignatureOwner)
            if (self.SignatureData is None):
                s += 'NONE'
            else:
                sdl = self.SignatureData
                for index in range(len(sdl)):
                    s += '%02X' % sdl[index]
            print(s)

    def Write(self, fs):
        """Write the object to the filestream.

        Args:
            fs (BinaryIO): filestream

        Raises:
            (Exception): Invalid filestream
            (Exception): Invalid object
        """
        if (fs is None):
            raise Exception("Invalid File Output Stream")
        if (self.SignatureData is None):
            raise Exception("Invalid object")

        fs.write(self.SignatureOwner.bytes_le)
        fs.write(self.SignatureData)

    def GetBytes(self) -> bytes:
        """Return bytes array produced by Write()."""
        with io.BytesIO() as fs:
            self.Write(fs)
            return fs.getvalue()

    def GetTotalSize(self):
        """Returns the total size of the object."""
        return EfiSignatureDataEfiCertX509.STATIC_STRUCT_SIZE + self.SignatureDataSize


class EfiSignatureDataEfiCertSha256(object):
    """An object representing a EFI_SIGNATURE_DATA Structure for Sha256 Certs."""
    STATIC_STRUCT_SIZE = 16 + hashlib.sha256().digest_size  # has guid and array
    FIXED_SIZE = True

    def __init__(self,
                 decodefs: BinaryIO = None,
                 createfs: BinaryIO = None,
                 digest: bytes = None,
                 sigowner: uuid = None):
        """Inits the object.

        Args:
            decodefs (BinaryIO): a filestream object of binary content that is the structure encoded
            createfs (BinaryIO): a filestream object that is the DER encoded x509 cert
            digest (bytes): a bytes object that contains the hash value for new signature data
            sigowner (uuid.UUID): the uuid object of the signature owner guid

        Raises:
            (Exception): Invalid FileStream size
            (Exception): Invalid Parameters
        """
        if (decodefs is not None):
            self.PopulateFromFileStream(decodefs)
        elif (createfs is not None):
            # create a new one
            self.SignatureOwner = sigowner
            self.SignatureData = (hashlib.sha256(createfs.read()).digest())
        elif (digest is not None):
            digest_length = len(digest)
            if (digest_length != hashlib.sha256().digest_size):
                raise Exception("Invalid digest length (found / expected): (%d / %d)",
                                digest_length, hashlib.sha256().digest_size)
            self.SignatureOwner = sigowner
            self.SignatureData = digest
        else:
            raise Exception("Invalid Parameters - Not Supported")

    def PopulateFromFileStream(self, fs):
        """Loads an object from a filestream.

        Args:
            fs (BinaryIO): filestream

        Raises:
            (Exception): Invalid filestream
        """
        if (fs is None):
            raise Exception("Invalid File Steam")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if ((end - offset) < EfiSignatureDataEfiCertSha256.STATIC_STRUCT_SIZE):  # size of the  data
            raise Exception("Invalid file stream size")

        self.SignatureOwner = uuid.UUID(bytes_le=fs.read(16))

        self.SignatureData = fs.read(hashlib.sha256().digest_size)

    def Print(self, compact: bool = False):
        """Prints to the console."""
        if not compact:
            print("EfiSignatureData - EfiSignatureDataEfiCertSha256")
            print("  Signature Owner:      %s" % str(self.SignatureOwner))
            print("  Signature Data: ", end="")
            if (self.SignatureData is None):
                print(" NONE")
            else:
                sdl = self.SignatureData
                for index in range(len(sdl)):
                    print("%02X" % sdl[index], end='')
                print("")
        else:
            s = 'ESD:EFI_CERT_SHA256,'
            s += "%s," % str(self.SignatureOwner)
            if (self.SignatureData is None):
                s += 'NONE'
            else:
                sdl = self.SignatureData
                for index in range(len(sdl)):
                    s += '%02X' % sdl[index]
            print(s)

    def Write(self, fs):
        """Write the object to the filestream.

        Args:
            fs (BinaryIO): filestream

        Raises:
            (Exception): Invalid filestream
            (Exception): Invalid object
        """
        if (fs is None):
            raise Exception("Invalid File Output Stream")
        if (self.SignatureData is None):
            raise Exception("Invalid object")

        fs.write(self.SignatureOwner.bytes_le)
        fs.write(self.SignatureData)

    def GetBytes(self) -> bytes:
        """Return bytes array produced by Write()."""
        with io.BytesIO() as fs:
            self.Write(fs)
            return fs.getvalue()

    def GetTotalSize(self):
        """Returns the total size of the object."""
        return EfiSignatureDataEfiCertSha256.STATIC_STRUCT_SIZE


class EfiSignatureHeader(object): # noqa
    def __init__(self): # noqa
        raise Exception("Not Implemented")


class EfiSignatureDataFactory(object):
    """Factory class for generating an EFI Signature struct."""
    EFI_CERT_SHA256_GUID = uuid.UUID('c1c41626-504c-4092-aca9-41f936934328')
    # EFI_CERT_RSA2048_GUID = uuid.UUID("0x3c5766e8, 0x269c, 0x4e34, 0xaa, 0x14, 0xed, 0x77, 0x6e, 0x85, 0xb3, 0xb6")
    # EFI_CERT_RSA2048_SHA256_GUID = uuid.UUID("0xe2b36190, 0x879b, 0x4a3d, 0xad, 0x8d, 0xf2, 0xe7, 0xbb, 0xa3, 0x27, 0x84")  # noqa: E501
    # EFI_CERT_SHA1_GUID = uuid.UUID("0x826ca512, 0xcf10, 0x4ac9, 0xb1, 0x87, 0xbe, 0x1, 0x49, 0x66, 0x31, 0xbd")
    # EFI_CERT_RSA2048_SHA1_GUID = uuid.UUID("0x67f8444f, 0x8743, 0x48f1, 0xa3, 0x28, 0x1e, 0xaa, 0xb8, 0x73, 0x60, 0x80")    # noqa: E501
    EFI_CERT_X509_GUID = uuid.UUID("a5c059a1-94e4-4aa7-87b5-ab155c2bf072")
    # EFI_CERT_SHA224_GUID = uuid.UUID("0xb6e5233, 0xa65c, 0x44c9, 0x94, 0x7, 0xd9, 0xab, 0x83, 0xbf, 0xc8, 0xbd")
    # EFI_CERT_SHA384_GUID = uuid.UUID("0xff3e5307, 0x9fd0, 0x48c9, 0x85, 0xf1, 0x8a, 0xd5, 0x6c, 0x70, 0x1e, 0x1")
    # EFI_CERT_SHA512_GUID = uuid.UUID("0x93e0fae, 0xa6c4, 0x4f50, 0x9f, 0x1b, 0xd4, 0x1e, 0x2b, 0x89, 0xc1, 0x9a")
    EFI_CERT_X509_SHA256_GUID = uuid.UUID("3bd2a492-96c0-4079-b420-fcf98ef103ed")
    # EFI_CERT_X509_SHA384_GUID = uuid.UUID("0x7076876e, 0x80c2, 0x4ee6, 0xaa, 0xd2, 0x28, 0xb3, 0x49, 0xa6, 0x86, 0x5b")     # noqa: E501
    # EFI_CERT_X509_SHA512_GUID = uuid.UUID("0x446dbf63, 0x2502, 0x4cda, 0xbc, 0xfa, 0x24, 0x65, 0xd2, 0xb0, 0xfe, 0x9d")     # noqa: E501
    # EFI_CERT_TYPE_PKCS7_GUID = uuid.UUID("0x4aafd29d, 0x68df, 0x49ee, 0x8a, 0xa9, 0x34, 0x7d, 0x37, 0x56, 0x65, 0xa7")

    @staticmethod
    def Factory(fs: BinaryIO, type, size):
        """This method is a factory for creating the correct Efi Signature Data object from the filestream.

        Uses a Filestream of an existing auth payload.

        Args:
            fs (BinaryIO): filestream to read
            type (uuid.UUID): Guid of the type
            size (int): decodesize for x509, struct size for Sha256
        """
        if (fs is None):
            raise Exception("Invalid File stream")

        if (type == EfiSignatureDataFactory.EFI_CERT_SHA256_GUID):
            if (size != EfiSignatureDataEfiCertSha256.STATIC_STRUCT_SIZE):
                raise Exception("Invalid Size 0x%x" % size)
            return EfiSignatureDataEfiCertSha256(decodefs=fs)

        elif (type == EfiSignatureDataFactory.EFI_CERT_X509_GUID):
            return EfiSignatureDataEfiCertX509(decodefs=fs, decodesize=size)

        else:
            logging.error("GuidType Value: %s" % type)
            raise Exception("Not Supported")
            return None

    @staticmethod
    def Create(type, ContentFileStream, sigowner):
        """Create a new EFI Sginature Data Object.

        Args:
            type (uuid.UUID): Guid of the type
            ContentFileStream (BinaryIO): filestream to read
            sigowner (uuid.UUID): the uuid object of the signature owner guid
        """
        if (ContentFileStream is None):
            raise Exception("Invalid Content File Stream")

        if (type == EfiSignatureDataFactory.EFI_CERT_SHA256_GUID):
            return EfiSignatureDataEfiCertSha256(createfs=ContentFileStream, sigowner=sigowner)
        elif (type == EfiSignatureDataFactory.EFI_CERT_X509_GUID):
            return EfiSignatureDataEfiCertX509(createfs=ContentFileStream, sigowner=sigowner)
        else:
            raise Exception("Not Supported")


class EfiSignatureList(object):
    """An object representing a EFI_SIGNATURE_LIST."""
    STATIC_STRUCT_SIZE = 16 + 4 + 4 + 4

    def __init__(self, filestream: BinaryIO = None, typeguid: uuid.UUID = None):
        """Inits the object.

        Args:
            filestream (:obj:`BinaryIO`, optional): a filestream object of binary content that is the structure encoded
            typeguid (:obj:`uuid.UUID`, optional): type guid
        """
        if (filestream is None):

            # Type of the signature. GUID signature types are defined in below.
            self.SignatureType = typeguid

            # Total size of the signature list, including this header.
            self.SignatureListSize = EfiSignatureList.STATIC_STRUCT_SIZE

            # Size of the signature header which precedes the array of signatures.
            self.SignatureHeaderSize = -1

            # Size of each signature.
            self.SignatureSize = 0

            # Header before the array of signatures. The format of this header is specified by the SignatureType.
            self.SignatureHeader = None

            # An array of signatures. Each signature is SignatureSize bytes in length.
            self.SignatureData_List = None

        else:
            self.PopulateFromFileStream(filestream)

    def PopulateFromFileStream(self, fs: BinaryIO):
        """Decodes a filestream and generates the structure.

        Args:
            fs (BinaryIO): filestream to load from

        Raises:
            (Exception): Invalid filestream
            (Exception): Invalid filestream size
            (Exception): Invalid siglist
            (Exception): Invalid parsing
        """
        if (fs is None):
            raise Exception("Invalid File Steam")

        # only populate from file stream those parts that are complete in the file stream
        start = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(start)

        if ((end - start) < EfiSignatureList.STATIC_STRUCT_SIZE):  # size of the static header data
            raise Exception("Invalid file stream size")

        self.SignatureType = uuid.UUID(bytes_le=fs.read(16))
        self.SignatureListSize = struct.unpack("<I", fs.read(4))[0]
        self.SignatureHeaderSize = struct.unpack("<I", fs.read(4))[0]
        self.SignatureSize = struct.unpack("<I", fs.read(4))[0]

        # check the total size of this is within the File
        if ((end - start) < self.SignatureListSize):
            logging.debug("SignatureListSize 0x%x" % self.SignatureListSize)
            logging.debug("End - Start is 0x%x" % (end - start))
            raise Exception("Invalid File Stream.  Not enough file content to cover the Sig List Size")

        # check that structure is built correctly and there is room within the structure total size to read the header
        if ((self.SignatureListSize - (fs.tell() - start)) < self.SignatureHeaderSize):
            raise Exception("Invalid Sig List.  Sizes not correct.  "
                            "SignatureHeaderSize extends beyond end of structure")

        # Signature Header is allowed to be nothing (size 0)
        self.SignatureHeader = None
        if (self.SignatureHeaderSize > 0):
            self.SignatureHeader = EfiSignatureHeader(fs, self.SignatureHeaderSize)

        if (((self.SignatureListSize - (fs.tell() - start)) % self.SignatureSize) != 0):
            raise Exception("Invalid Sig List.  Signature Data Array is not a valid size")

        self.SignatureData_List = []
        while ((start + self.SignatureListSize) > fs.tell()):
            # double check that everything is adding up correctly.
            if ((start + self.SignatureListSize - fs.tell() - self.SignatureSize) < 0):
                raise Exception("Invalid Signature List Processing.  Signature Data not correctly parsed!!")
            a = EfiSignatureDataFactory.Factory(fs, self.SignatureType, self.SignatureSize)
            self.SignatureData_List.append(a)

    def Print(self, compact: bool = False):
        """Prints to the console."""
        if not compact:
            print("EfiSignatureList")
            print("  Signature Type:        %s" % str(self.SignatureType))
            print("  Signature List Size:   0x%x" % self.SignatureListSize)
            print("  Signature Header Size: 0x%x" % self.SignatureHeaderSize)
            print("  Signature Size:        0x%x" % self.SignatureSize)
            if (self.SignatureHeader is not None):
                self.SignatureHeader.Print(compact=compact)
            else:
                print("  Signature Header:      NONE")
        else:
            csv = "ESL:"
            csv += "%s" % str(self.SignatureType)
            csv += ",0x%x" % self.SignatureListSize
            csv += ",0x%x" % self.SignatureHeaderSize
            csv += ",0x%x" % self.SignatureSize
            if (self.SignatureHeader is not None):
                csv += self.SignatureHeader.Print(compact=compact)
            else:
                csv += ",NONE"
            print(csv)

        if (self.SignatureData_List is not None):
            for a in self.SignatureData_List:
                a.Print(compact=compact)

    def Write(self, fs):
        """Serializes the object and writes it to a filestream.

        Raises:
            (Exception): Invalid filestream
            (Exception): Uninitialized Sig header
        """
        if (fs is None):
            raise Exception("Invalid File Output Stream")
        if ((self.SignatureHeader is None) and (self.SignatureHeaderSize == -1)):
            raise Exception("Invalid object.  Uninitialized Sig Header")

        fs.write(self.SignatureType.bytes_le)
        fs.write(struct.pack("<I", self.SignatureListSize))
        fs.write(struct.pack("<I", self.SignatureHeaderSize))
        fs.write(struct.pack("<I", self.SignatureSize))
        if (self.SignatureHeader is not None):
            self.SignatureHeader.Write(fs)

        if (self.SignatureData_List is not None):
            for a in self.SignatureData_List:
                a.Write(fs)

    def GetBytes(self) -> bytes:
        """Return bytes array produced by Write()."""
        with io.BytesIO() as fs:
            self.Write(fs)
            return fs.getvalue()

    def AddSignatureHeader(self, SigHeader, SigSize=0):
        """Adds the Signature Header.

        Raises:
            (Exception): Signature header already set
        """
        if (self.SignatureHeader is not None):
            raise Exception("Signature Header already set")

        if (self.SignatureHeaderSize != -1):
            raise Exception("Signature Header already set (size)")

        if (self.SignatureSize != 0):
            raise Exception("Signature Size already set")

        if (self.SignatureData_List is not None):
            raise Exception("Signature Data List is already initialized")

        self.SignatureHeader = SigHeader
        if (SigHeader is None):
            self.SignatureHeaderSize = 0
            self.SignatureSize = SigSize
        else:
            self.SignatureHeaderSize = SigHeader.GetTotalSize()
            self.SignatureSize = SigHeader.GetSizeOfSignatureDataEntry()
            self.SignatureListSize += self.SignatureHeaderSize

    def AddSignatureData(self, SigDataObject):
        """Adds the Signature Data.

        Raises:
            (Exception): Signature size does not match Sig Data Object size
        """
        if (self.SignatureSize == 0):
            raise Exception("Before adding Signature Data you must have set the Signature Size")

        if (self.SignatureSize != SigDataObject.GetTotalSize()):
            raise Exception("Can't add Signature Data of different size")

        if (self.SignatureData_List is None):
            self.SignatureData_List = []

        self.SignatureData_List.append(SigDataObject)
        self.SignatureListSize += self.SignatureSize

    def MergeSignatureList(self, esl):
        """Add the EfiSignatureData entries within the supplied EfiSignatureList to the current object.

        No sorting or deduplication is performed, the EfiSignatureData elements are simply appended
        """
        if not isinstance(esl, EfiSignatureList):
            raise Exception("Parameter 1 'esl' must be of type EfiSignatureList")

        if (self.SignatureType != esl.SignatureType):
            raise Exception("Signature Types must match")

        if (self.SignatureHeaderSize > 0
           or esl.SignatureHeaderSize > 0):
            raise Exception("Merge does not support Signature Headers")
        self.SignatureHeaderSize = 0

        if (esl.SignatureListSize == EfiSignatureList.STATIC_STRUCT_SIZE):
            # supplied EfiSignatureList is empty, return
            return

        FixedSizeData = esl.SignatureData_List[0].FIXED_SIZE
        if not FixedSizeData:
            raise Exception("Can only merge EfiSignatureLists with fixed-size data elements")

        if (self.SignatureData_List is None):
            self.SignatureData_List = []
            self.SignatureSize = esl.SignatureSize

        self.SignatureData_List += esl.SignatureData_List
        self.SignatureListSize += esl.SignatureListSize - EfiSignatureList.STATIC_STRUCT_SIZE

    def SortBySignatureDataValue(self, deduplicate: bool = True):
        """Sort self's SignatureData_List by SignatureData values (ignores SigOwner) & optionally deduplicate.

        When deduplicate is true, remove duplicate SignatureData values from self and return them in an
        EfiSignatureList.  This EfiSignatureList of duplicates is itself not deduplicated.

        When deduplicate is false, returns an empty EfiSignatureList (has 0 Data elements)
        """
        # initialize the duplicate list, an EFI_SIGNATURE_LIST with no signature data entries
        dupes = EfiSignatureList(typeguid=self.SignatureType)
        dupes.SignatureListSize = EfiSignatureList.STATIC_STRUCT_SIZE
        dupes.SignatureHeaderSize = self.SignatureHeaderSize
        dupes.SignatureHeader = self.SignatureHeader
        dupes.SignatureSize = 0
        dupes.SignatureData_List = []

        # if nothing to sort, return the empty dupe list
        if (self.SignatureData_List is None
           or len(self.SignatureData_List) == 1):
            return dupes

        self.SignatureData_List.sort(key=attrgetter('SignatureData'))

        if (deduplicate is False):
            return dupes  # return empty dupe list without performing deduplicate

        # perform deduplicate on self
        last = len(self.SignatureData_List) - 1  # index of the last item
        for i in range(last - 1, -1, -1):
            if self.SignatureData_List[last].SignatureData == self.SignatureData_List[i].SignatureData:
                dupes.SignatureData_List.insert(0, self.SignatureData_List[last])
                dupes.SignatureListSize += self.SignatureSize
                del self.SignatureData_List[last]
                self.SignatureListSize -= self.SignatureSize
                last = i
            else:
                last = i

        # only initialize dupes.SignatureSize if duplicate elements are present
        if (len(dupes.SignatureData_List) > 0):
            dupes.SignatureSize = self.SignatureSize

        return dupes


class EfiSignatureDatabase(object):
    """Concatenation of EFI_SIGNATURE_LISTs, as is returned from UEFI GetVariable() on PK, KEK, db, & dbx.

    Useful for parsing and building the contents of the Secure Boot variables
    """

    def __init__(self, filestream: BinaryIO = None, EslList: List[EfiSignatureList] = None):
        """Inits an Efi Signature Database object.

        Args:
            filestream (:obj:`BinaryIO`, optional): Inits the object with this stream
            EslList (:obj:`list[EfiSignatureList]`, optional): Inits the object with this list
        """
        if filestream:
            self.EslList = []
            self.PopulateFromFileStream(filestream)
        else:
            self.EslList = [] if EslList is None else EslList

    def PopulateFromFileStream(self, fs: BinaryIO):
        """Decodes a filestream and generates the structure.

        Args:
            fs (BinaryIO): filestream to load from

        Raises:
            (Exception): Invalid filestream
            (Exception): Invalid filestream size
        """
        if (fs is None):
            raise Exception("Invalid File Stream")
        self.EslList = []
        begin = fs.tell()
        fs.seek(0, io.SEEK_END)
        end = fs.tell()  # end is offset after last byte
        fs.seek(begin)
        while (fs.tell() != end):
            Esl = EfiSignatureList(fs)
            self.EslList.append(Esl)

    def Print(self, compact: bool = False):
        """Prints to the console."""
        for Esl in self.EslList:
            Esl.Print(compact=compact)

    def Write(self, fs):
        """Serializes the object and writes it to a filestream.

        Raises:
            (Exception): Invalid filestream
        """
        if (fs is None):
            raise Exception("Invalid File Output Stream")
        for Esl in self.EslList:
            Esl.Write(fs)

    def GetBytes(self) -> bytes:
        """Return bytes array produced by Write()."""
        with io.BytesIO() as fs:
            self.Write(fs)
            return fs.getvalue()

    def GetCanonicalAndDupes(self):
        """Compute and return a tuple containing both a canonicalized database & a database of duplicates.

        Returns:
            (Tuple[EfiSignatureDatabase, EfiSignatureDatabase]): (canonical, duplicates)

        NOTE:
        canonical is an EfiSignatureDatabase where EfiSignatureLists are merged (where possible),
            deduplicated, & sorted, and the EfiSignatureData elements are also deduplicated & sorted
        duplicates is an EfiSignatureDatabase with EfiSignatureLists containing any duplicated
            EfiSignatureData entries (only the data contents are checked for effective equality,
            signature owner is ignored)
        """
        # First group EFI_SIGNATURE_LISTS by type, merging them where possible

        sha256esl = None
        x509eslList = None

        for Esl in self.EslList:
            if (Esl.SignatureData_List is None):  # discard empty EfiSignatureLists
                continue
            if (Esl.SignatureType == EfiSignatureDataFactory.EFI_CERT_SHA256_GUID):
                if (sha256esl is None):
                    sha256esl = Esl  # initialize it
                else:
                    sha256esl.MergeSignatureList(Esl)
            elif (Esl.SignatureType == EfiSignatureDataFactory.EFI_CERT_X509_GUID):
                if (x509eslList is None):
                    x509eslList = [Esl]  # initialize it
                else:
                    x509eslList.append(Esl)
            else:
                raise Exception("Unsupported signature type %s", Esl.SignatureType)

        # for each type, sort and de-duplicate, and then populate the respective databases
        # note the ordering of this section is the prescribed canonical order
        # first 1 EfiSignatureList for SHA256 hashes, EfiSignatureData elements sorted ascending
        # followed by EfiSignatureLists for each x509 certificate, sorted ascending by data content

        canonicalDb = EfiSignatureDatabase()
        duplicatesDb = EfiSignatureDatabase()

        if (sha256esl is not None):
            dupes = sha256esl.SortBySignatureDataValue(deduplicate=True)
            canonicalDb.EslList.append(sha256esl)
            duplicatesDb.EslList.append(dupes)

        if (x509eslList is not None):
            x509eslList.sort(key=attrgetter('SignatureData_List'))
            for esl in x509eslList:
                if not canonicalDb.EslList:
                    canonicalDb.EslList.append(esl)
                elif esl == canonicalDb.EslList[-1]:
                    duplicatesDb.EslList.append(esl)
                else:
                    canonicalDb.EslList.append(esl)

        return (canonicalDb, duplicatesDb)

    def GetCanonical(self):
        """Return a canonicalized EfiSignatureDatabase, see GetCanonicalAndDupes() for more details."""
        (canonical, dupes) = self.GetCanonicalAndDupes()
        return canonical

    def GetDuplicates(self):
        """Return an EfiSignatureDatabase of duplicates, see GetCanonicalAndDupes() for more details."""
        (canonical, dupes) = self.GetCanonicalAndDupes()
        return dupes


class EfiTime(object):
    """Object representing an EFI_TIME."""
    STATIC_STRUCT_SIZE = 16

    def __init__(self, Time=datetime.datetime.now(), decodefs=None):
        """Inits an EFI_TIME object.

        Args:
            Time (:obj:`datetime`, optional): Inits object with specified date (if decodefs not set)
            decodefs (:obj:`BinaryIO`, optional): Inits the object with this stream
        """
        if (decodefs is None):
            self.Time = Time
        else:
            self.PopulateFromFileStream(decodefs)

    def PopulateFromFileStream(self, fs):
        """Decodes a filestream and generates the structure.

        Args:
            fs (BinaryIO): filestream to load from

        Raises:
            (Exception): Invalid filestream
            (Exception): Invalid filestream size
        """
        if (fs is None):
            raise Exception("Invalid File Steam")

        # only populate from file stream those parts that are complete in the file stream
        start = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(start)

        if ((end - start) < EfiTime.STATIC_STRUCT_SIZE):  # size of the static structure data
            raise Exception("Invalid file stream size")

        Year = struct.unpack("<H", fs.read(2))[0]
        Month = struct.unpack("<B", fs.read(1))[0]
        Day = struct.unpack("<B", fs.read(1))[0]
        Hour = struct.unpack("<B", fs.read(1))[0]
        Minute = struct.unpack("<B", fs.read(1))[0]
        Second = struct.unpack("<B", fs.read(1))[0]
        fs.seek(1, 1)  # seek past pad1
        NanoSecond = struct.unpack("<I", fs.read(4))[0]
        TimeZone = struct.unpack("<h", fs.read(2))[0]
        Daylight = struct.unpack("<B", fs.read(1))[0]
        fs.seek(1, 1)  # seek past pad2

        self.Time = datetime.datetime(Year, Month, Day, Hour, Minute, Second, NanoSecond // 1000)
        logging.debug("I don't know how to deal with TimeZone or Daylight and I don't care at the moment")
        logging.debug("Timezone value is: 0x%x" % TimeZone)
        logging.debug("Daylight value is: 0x%X" % Daylight)

    def Print(self):
        """Prints to the console."""
        print("EfiTime: %s" % datetime.datetime.strftime(self.Time, "%A, %B %d, %Y %I:%M%p"))

    def Write(self, fs):
        """Serializes the object and writes it to a filestream.

        Raises:
            (Exception): Invalid filestream
        """
        if (fs is None):
            raise Exception("Invalid File Output Stream")

        fs.write(struct.pack("<H", self.Time.year))
        fs.write(struct.pack("<B", self.Time.month))
        fs.write(struct.pack("<B", self.Time.day))
        fs.write(struct.pack("<B", self.Time.hour))
        fs.write(struct.pack("<B", self.Time.minute))
        fs.write(struct.pack("<B", self.Time.second))
        fs.write(struct.pack("<B", 0))  # Pad1
        fs.write(struct.pack("<I", 0))  # Nano Seconds
        fs.write(struct.pack("<h", 0))  # TimeZone
        fs.write(struct.pack("<B", 0))  # Daylight
        fs.write(struct.pack("<B", 0))  # Pad2


class EFiVariableAuthentication2(object):
    """An object representing a EFI_VARIABLE_AUTHENTICATION_2."""
    def __init__(self, Time=datetime.datetime.now(), decodefs=None):
        """Inits an EFI_VARIABLE_AUTHENTICATION_2 object.

        Args:
            Time (:obj:`datetime`, optional): Inits object with specified date (if decodefs not set)
            decodefs (:obj:`BinaryIO`, optional): Inits the object with this stream
        """
        if (decodefs is None):
            self.EfiTime = EfiTime(Time=Time)
            self.AuthInfo = WinCertUefiGuid()
            self.Payload = None
            self.PayloadSize = 0
            self.SigListPayload = None
        else:
            self.PopulateFromFileStream(decodefs)

    def PopulateFromFileStream(self, fs):
        """Decodes a filestream and generates the structure.

        Args:
            fs (BinaryIO): filestream to load from

        Raises:
            (Exception): Invalid filestream
        """
        if (fs is None):
            raise Exception("Invalid File Steam")
        self.EfiTime = EfiTime(decodefs=fs)
        self.AuthInfo = WinCert.Factory(fs)
        self.Payload = None
        self.SigListPayload = None

        self.SetPayload(fs)

    def Print(self):
        """Prints to the console."""
        print("EFiVariableAuthentication2")
        self.EfiTime.Print()
        self.AuthInfo.Print()
        print("-------------------- VARIABLE PAYLOAD --------------------")
        if (self.SigListPayload is not None):
            self.SigListPayload.Print()

        elif (self.Payload is not None):
            print("Raw Data: ")
            sdl = self.Payload.tolist()
            if (self.PayloadSize != len(sdl)):
                raise Exception("Invalid Payload Data Size vs Length of data")
            PrintByteList(sdl)

    def Write(self, fs):
        """Serializes the object and writes it to a filestream.

        Raises:
            (Exception): Invalid filestream
        """
        if (fs is None):
            raise Exception("Invalid File Output Stream")

        self.EfiTime.Write(fs)
        self.AuthInfo.Write(fs)
        if (self.Payload is not None):
            fs.write(self.Payload)

    def SetPayload(self, fs):
        """Decodes a filestream and generates the payload.

        Args:
            fs (BinaryIO): filestream to load from

        Raises:
            (Exception): Invalid filestream
        """
        if (fs is None):
            raise Exception("Invalid File Input Stream")

        # Find the payload size
        start = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(start)
        self.PayloadSize = end - start
        if (self.PayloadSize == 0):
            logging.debug("No Payload for this EfiVariableAuthenticated2 Object")
            return

        # read as siglist
        try:
            self.SigListPayload = EfiSignatureList(fs)
        except Exception as e:
            logging.debug("Exception Trying to parse SigList Payload.  \n%s" % str(e))

        # reset the file pointer
        fs.seek(start)
        self.Payload = memoryview(fs.read(self.PayloadSize))


'''
THESE ARE NOT SUPPORTED IN THE TOOL

```
typedef struct {
  ///
  /// The SHA256 hash of an X.509 certificate's To-Be-Signed contents.
  ///
  EFI_SHA256_HASH     ToBeSignedHash;
  ///
  /// The time that the certificate shall be considered to be revoked.
  ///
  EFI_TIME            TimeOfRevocation;
} EFI_CERT_X509_SHA256;
```

```
typedef struct {
  ///
  /// The SHA384 hash of an X.509 certificate's To-Be-Signed contents.
  ///
  EFI_SHA384_HASH     ToBeSignedHash;
  ///
  /// The time that the certificate shall be considered to be revoked.
  ///
  EFI_TIME            TimeOfRevocation;
} EFI_CERT_X509_SHA384;
```

```
typedef struct {
  ///
  /// The SHA512 hash of an X.509 certificate's To-Be-Signed contents.
  ///
  EFI_SHA512_HASH     ToBeSignedHash;
  ///
  /// The time that the certificate shall be considered to be revoked.
  ///
  EFI_TIME            TimeOfRevocation;
} EFI_CERT_X509_SHA512;
```
'''



# @file
# Code to test UEFI status code module
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
from edk2toollib.uefi.status_codes import UefiStatusCode


class TestUefiStatusCodes (unittest.TestCase):

    def test_Hex64ToString_NotError(self):
        StatusCode = "0x0000000000000000"
        self.assertEqual(UefiStatusCode().ConvertHexString64ToString(StatusCode), "Success")

    def test_Hex64ToString_ErrorNotFound(self):
        StatusCode = "0x800000000000000E"
        self.assertEqual(UefiStatusCode().ConvertHexString64ToString(StatusCode), "Not Found")

    def test_Hex64ToString_Error_Invalid_Len(self):
        StatusCode = hex(len(UefiStatusCode.ErrorCodeStrings) + 0x8000000000000000)
        self.assertEqual(UefiStatusCode().ConvertHexString64ToString(StatusCode), "Undefined StatusCode")

    def test_Hex32ToString_NotError(self):
        StatusCode = "0x00000000"
        self.assertEqual(UefiStatusCode().ConvertHexString32ToString(StatusCode), "Success")

    def test_Hex32ToString_ErrorInvalidParameter(self):
        StatusCode = "0x80000002"
        self.assertEqual(UefiStatusCode().ConvertHexString32ToString(StatusCode), "Invalid Parameter")

    def test_Hex32ToString_Error_Invalid_Len(self):
        StatusCode = hex(len(UefiStatusCode.ErrorCodeStrings) + 0x80000000)
        self.assertEqual(UefiStatusCode().ConvertHexString32ToString(StatusCode), "Undefined StatusCode")

    def test_Hex32ToString_NonError_Invalid_Len(self):
        StatusCode = hex(len(UefiStatusCode.NonErrorCodeStrings))
        self.assertEqual(UefiStatusCode().ConvertHexString32ToString(StatusCode), "Undefined StatusCode")



# @file gitignore_parser_test.py
# unit test for gitignore_parser module.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
import tempfile
import edk2toollib.gitignore_parser as gitignore_parser
import os


def BuildGitIgnore(root):
    path = os.path.join(root, ".gitignore")
    with open(path, "w") as gitignore:
        gitignore.write("/Build/\n")
        gitignore.write("./Logs/\n")
        gitignore.write("**/Test/\n")
        gitignore.write(".DS_Store\n")
        gitignore.write("*_extdep/\n")
        gitignore.write("*.pyc\n")
        gitignore.write("__pycache__/\n")
        gitignore.write("tags/\n")
        gitignore.write(".vscode/\n")
        gitignore.write("*.bak\n")
        gitignore.write("BuildConfig.conf\n")
        gitignore.write(" \n")
        gitignore.write("*.exe\n")
        gitignore.write("!important.exe\n")
        gitignore.write("/BaseTools/BaseToolsBuild/\n")
        gitignore.write("/docs/**/*.txt\n")
        gitignore.write("/reader/*\n")
        gitignore.write("out?.log\n")
        gitignore.write("log[0-9].txt\n")
        gitignore.write("log0[!a-z].txt\n")
    return path


class GitIgnoreParserTest(unittest.TestCase):

    def test_gitignoreparser_filter(self):
        '''
        Ensure the gitignore parser filters files and folders correctly per
        what is specified in the parsed .gitignore file.
        '''
        with tempfile.TemporaryDirectory() as root:

            gitignore_path = BuildGitIgnore(root)
            rule_tester = gitignore_parser.parse_gitignore_file(gitignore_path)

            # Test a rule which specifies that a specific folder at root is
            # correctly filtered.

            # Example line in a .gitignore
            # /Build/
            self.assertTrue(rule_tester(os.path.join(root, "Build")))
            self.assertFalse(rule_tester(os.path.join(root, "T", "Build")))

            # Test a rule which specifies that a specific folder is not
            # filtered, but the contents of the folder is filtered

            # Example ine in a .gitignore
            # /reader/*
            self.assertFalse(rule_tester(os.path.join(root, "reader")))
            self.assertTrue(rule_tester(os.path.join(root, "reader", "testing.txt")))

            # Test a rule which specifies that a folder at any depth from the
            # root is correclty filtered.

            # Example line in a .gitignore
            # **/Test/
            self.assertTrue(rule_tester(os.path.join(root, "Test")))
            self.assertTrue(rule_tester(os.path.join(root, "T", "Test")))

            # Test a rule which specifies that a folder containing a specific
            # string at any depth is correctly filtered.

            # Example line in a .gitignore
            # *_extdep
            self.assertTrue(rule_tester(os.path.join(root, 'test_extdep')))
            self.assertTrue(rule_tester(os.path.join(root, 'T', 'test_extdep')))

            # Test a rule which specifies that a file of a specific type at any
            # depth is correctly filtered.

            # Example line in a .gitignore
            # .DS_Store
            self.assertTrue(rule_tester(os.path.join(root, "file.DS_Store")))
            self.assertTrue(rule_tester(os.path.join(root, "T", "file.DS_Store")))

            # Test a rule which specifies an exception for a previous rule is
            # correctly filtered.

            # Example line in a .gitignore
            # *.exe
            # !important.exe
            self.assertTrue(rule_tester(os.path.join(root, 'test.exe')))
            self.assertTrue(rule_tester(os.path.join(root, 'T', 'test.exe')))
            self.assertFalse(rule_tester(os.path.join(root, 'important.exe')))
            self.assertFalse(rule_tester(os.path.join(root, 'T', 'important.exe')))

            # Test a rule which specifies a file type in any directory under a
            # specific folder are correctly filtered.

            # Example line in a .gitignore
            # /docs/**/*.txt
            self.assertTrue(rule_tester(os.path.join(root, 'docs', 'notes.txt')))
            self.assertFalse(rule_tester(os.path.join(root, 'docs', 'Readme.md')))
            self.assertTrue(rule_tester(os.path.join(root, 'docs', 'developing', 'notes.txt')))
            self.assertFalse(rule_tester(os.path.join(root, 'docs', 'developing', 'Readme.md')))

            # Test a rule which specifies a file with a specific naming convention be
            # correctly filtered

            # Example line in a .gitignore
            # out?.log

            self.assertTrue(rule_tester(os.path.join(root, 'out1.log')))
            self.assertTrue(rule_tester(os.path.join(root, 'outF.log')))
            self.assertFalse(rule_tester(os.path.join(root, 'out11.log')))

            # log[0-9].txt
            self.assertTrue(rule_tester(os.path.join(root, 'log1.txt')))
            self.assertFalse(rule_tester(os.path.join(root, 'logF.txt')))
            self.assertFalse(rule_tester(os.path.join(root, 'log11.txt')))

            # log0[!a-z].txt
            self.assertTrue(rule_tester(os.path.join(root, 'log01.txt')))
            self.assertFalse(rule_tester(os.path.join(root, 'log0a.txt')))
            self.assertTrue(rule_tester(os.path.join(root, 'log0A.txt')))

    def test_rule_from_pattern(self):

        # Test bad basepath
        self.assertRaises(ValueError, gitignore_parser.rule_from_pattern, "", "Test")

        # Test ignoring comments, line separators, and incorrect astrick count
        with tempfile.TemporaryDirectory() as root:
            self.assertIsNone(gitignore_parser.rule_from_pattern("# Comment", root))
            self.assertIsNone(gitignore_parser.rule_from_pattern(" ", root))
            self.assertIsNone(gitignore_parser.rule_from_pattern("***", root))
